from collections import defaultdict

from .models import Discussion
from .compat import User
from .utils import get_profile_model, queryset_to_dict


def lookup_discussions(recipients):
    recipients_by_ids = queryset_to_dict(
        recipients, key="discussion_id", singular=False
    )

    discussions = (
        Discussion.objects.filter(pk__in=recipients_by_ids.keys())
        .select_related("sender")
        .prefetch_related("recipients")
    )

    for discussion in discussions:
        if discussion.pk in recipients_by_ids:
            for recipient in recipients_by_ids[discussion.pk]:
                recipient.discussion = discussion

    return recipients


def lookup_profiles(recipients):
    user_ids = defaultdict(list)

    for recipient in recipients:
        try:
            user_ids[recipient.user_id].append(recipient.user)
            user_ids[recipient.discussion.sender_id].append(recipient.discussion.sender)
        except User.DoesNotExist:
            pass

        for user in recipient.discussion.recipients.all():
            user_ids[user.pk].append(user)

    Profile = get_profile_model()

    if Profile:
        profiles = queryset_to_dict(
            Profile.objects.filter(user__in=user_ids), key="user_id"
        )

        for user_id, profile in profiles.iteritems():
            if user_id in user_ids:
                for user in user_ids[user_id]:
                    profile.user = user
                    user._profile_cache = profile

    return recipients



from __future__ import unicode_literals

from django.test import TestCase

from ..models import Message, Recipient
from ..compat import truncate_words


class MessageModelTests(TestCase):
    fixtures = ["users.json", "messages.json"]

    def test_string_formatting(self):
        """ Test the human representation of a message """
        message = Message.objects.get(pk=1)
        truncated_body = truncate_words(message.body, 10)
        self.assertEqual("%s" % message, truncated_body)


class MessageRecipientModelTest(TestCase):
    fixtures = ["users.json", "messages.json"]

    def test_string_formatting(self):
        """ Test the human representation of a recipient """
        recipient = Recipient.objects.get(pk=1)

        valid_unicode = "%s" % recipient.discussion

        self.assertEqual("%s" % recipient, valid_unicode)

    def test_recipients_count(self):
        recipient = Recipient.objects.get(pk=1)

        discussion = recipient.discussion

        assert discussion.recipients_count == discussion.recipients.count()

    def test_new(self):
        """ Test if the message that is new is correct """
        new_message = Recipient.objects.get(pk=1)
        read_message = Recipient.objects.get(pk=2)

        self.assertFalse(new_message.is_read())
        self.assertTrue(read_message.is_read())



from django.conf.urls import url

from . import views, settings
from .utils import load_class

pre_filter = load_class(settings.PRE_FILTER)


urlpatterns = [
    url(
        r"^compose(?:\/(?P<recipients>[\w\-\_\+]+))?/$",
        pre_filter(views.MessageComposeView.as_view()),
        name="discussions_compose",
    ),
    url(
        r"^view/(?P<discussion_id>[\d]+)/$",
        pre_filter(views.DiscussionDetailView.as_view()),
        name="discussions_detail",
    ),
    url(
        r"^remove/(?:(?P<folder_id>[\d]+))?$",
        pre_filter(views.DiscussionRemoveView.as_view()),
        name="discussions_remove",
    ),
    url(
        r"^unremove/$",
        pre_filter(views.DiscussionRemoveView.as_view()),
        {"undo": True},
        name="discussions_unremove",
    ),
    url(
        r"^(?:(?P<username>[\w\-\_]+))?(?:folder/(?P<folder_id>[\d]+))?$",
        pre_filter(views.DiscussionListView.as_view()),
        name="discussions_list",
    ),
    url(
        r"^filter/sent/(?:(?P<folder_id>[\d]+))?$",
        pre_filter(views.status.DiscussionSentView.as_view()),
        name="discussions_sent",
    ),
    url(
        r"^filter/unread/(?:(?P<folder_id>[\d]+))?$",
        pre_filter(views.status.DiscussionUnreadView.as_view()),
        name="discussions_unread",
    ),
    url(
        r"^filter/read/(?:(?P<folder_id>[\d]+))?$",
        pre_filter(views.status.DiscussionReadView.as_view()),
        name="discussions_read",
    ),
    url(
        r"^filter/trash/(?:(?P<folder_id>[\d]+))?$",
        pre_filter(views.status.DiscussionDeletedView.as_view()),
        name="discussions_deleted",
    ),
    url(
        r"^leave/$",
        pre_filter(views.DiscussionLeaveView.as_view()),
        name="discussions_leave",
    ),
    url(
        r"^folders/all/$",
        pre_filter(views.FoldersListView.as_view()),
        name="discussions_folders_list",
    ),
    url(
        r"^move/(?:(?P<folder_id>[\d]+))?$",
        pre_filter(views.DiscussionMoveView.as_view()),
        name="discussions_move",
    ),
    url(
        r"^mark/read/(?:(?P<folder_id>[\d]+))?$",
        pre_filter(views.DiscussionMarkAsReadView.as_view()),
        name="discussions_mark_as_read",
    ),
    url(
        r"^mark/unread/(?:(?P<folder_id>[\d]+))?$",
        pre_filter(views.DiscussionMarkAsUnreadView.as_view()),
        name="discussions_mark_as_unread",
    ),
    url(
        r"^folder/create/$",
        pre_filter(views.FolderCreateView.as_view()),
        name="discussions_folder_create",
    ),
    url(
        r"^folder/(?P<folder_id>[\d]+)/update$",
        pre_filter(views.FolderUpdateView.as_view()),
        name="discussions_folder_update",
    ),
    url(
        r"^folder/remove/(?P<folder_id>[\d]+)/$",
        pre_filter(views.FolderRemoveView.as_view()),
        name="discussions_folder_remove",
    ),
]



from django import forms
from django.utils.translation import ugettext_lazy as _

from ..fields import CommaSeparatedUserField
from ..models import Discussion, Folder


class ComposeForm(forms.Form):
    to = CommaSeparatedUserField(label=_("To"))
    subject = forms.CharField(
        label=_("Subject"),
        widget=forms.TextInput({"placeholder": _("Write your title")}),
        required=True,
    )
    body = forms.CharField(
        label=_("Message"),
        widget=forms.Textarea(
            {"class": "message", "placeholder": _("Write your message")}
        ),
        required=True,
    )

    def save(self, sender):
        """
        Save the discussion and send it out into the wide world.

        :param sender:
            The :class:`User` that sends the message.

        :return: The saved :class:`Discussion`.

        """
        to_user_list = self.cleaned_data["to"]
        subject = self.cleaned_data["subject"]
        body = self.cleaned_data["body"]

        self.discussion = Discussion.objects.send_message(
            sender, to_user_list, subject, body
        )

        return self.discussion


class ReplyForm(forms.Form):
    body = forms.CharField(
        label=_("Message"),
        widget=forms.Textarea(
            {"class": "message", "placeholder": _("Write your message")}
        ),
        required=True,
    )

    def __init__(self, *args, **kwargs):
        self.discussion = kwargs.pop("discussion")

        super(ReplyForm, self).__init__(*args, **kwargs)

    def save(self, sender):
        return self.discussion.add_message(self.cleaned_data["body"], sender)


class FolderForm(forms.ModelForm):
    name = forms.CharField(
        label=_("Name"),
        widget=forms.TextInput({"placeholder": _("Write the folder's name")}),
        required=True,
    )

    class Meta:
        exclude = ("user", "discussions", "created_at")
        model = Folder

    def save(self, user):
        self.instance.user = user

        return super(FolderForm, self).save()



from django.db import models
from django.conf import settings
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import python_2_unicode_compatible

from discussions.managers import DiscussionManager, RecipientManager, MessageManager

from django.utils.timezone import now as tznow

from ..utils import get_model_string

from model_utils import Choices


AUTH_USER_MODEL = getattr(settings, "AUTH_USER_MODEL", "auth.User")


@python_2_unicode_compatible
class Recipient(models.Model):
    """
    Intermediate model to allow per recipient marking as
    deleted, read etc. of a message.

    """

    STATUS = Choices(
        (0, "read", _("read")), (1, "unread", _("unread")), (2, "deleted", _("deleted"))
    )

    user = models.ForeignKey(
        AUTH_USER_MODEL, on_delete=models.CASCADE, verbose_name=_("recipient")
    )

    discussion = models.ForeignKey(
        get_model_string("Discussion"),
        on_delete=models.CASCADE,
        verbose_name=_("discussion"),
    )

    folder = models.ForeignKey(
        get_model_string("Folder"),
        verbose_name=_("folder"),
        null=True,
        blank=True,
        related_name="recipients",
        on_delete=models.SET_NULL,
    )

    read_at = models.DateTimeField(_("read at"), null=True, blank=True)

    deleted_at = models.DateTimeField(_("recipient deleted at"), null=True, blank=True)

    status = models.PositiveSmallIntegerField(
        choices=STATUS, default=STATUS.unread, verbose_name=_("Status"), db_index=True
    )

    objects = RecipientManager()

    class Meta:
        verbose_name = _("recipient")
        verbose_name_plural = _("recipients")
        app_label = "discussions"
        abstract = True

    def __str__(self):
        return _("%(discussion)s") % {"discussion": self.discussion}

    def is_read(self):
        """ Returns a boolean whether the recipient has read the message """
        return self.status == self.STATUS.read

    def is_unread(self):
        """ Returns a boolean whether the recipient hasn't read the message """
        return self.status == self.STATUS.unread

    def is_deleted(self):
        """ Returns a boolean whether the recipient has deleted the message """
        return self.status == self.STATUS.deleted

    def mark_as_deleted(self, commit=True):
        self.deleted_at = tznow()
        self.status = self.STATUS.deleted

        if commit:
            self.save()

    def mark_as_read(self, commit=True):
        self.read_at = tznow()
        self.status = self.STATUS.read

        if commit:
            self.save()

    def mark_as_unread(self, commit=True):
        self.status = self.STATUS.unread

        if commit:
            self.save()


@python_2_unicode_compatible
class Discussion(models.Model):
    """ Private message model, from user to user(s) """

    sender = models.ForeignKey(
        AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="discussions_sent",
        verbose_name=_("sender"),
    )

    recipients = models.ManyToManyField(
        AUTH_USER_MODEL,
        through="Recipient",
        related_name="discussions_received",
        verbose_name=_("recipients"),
    )

    created_at = models.DateTimeField(_("created at"), auto_now_add=True)

    updated_at = models.DateTimeField(_("updated at"), null=True, blank=True)

    sender_deleted_at = models.DateTimeField(
        _("sender deleted at"), null=True, blank=True
    )

    latest_message = models.ForeignKey(
        get_model_string("Message"),
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="latest_discussions",
    )

    subject = models.CharField(max_length=255)

    recipients_count = models.PositiveIntegerField(default=0, null=True, blank=True)

    messages_count = models.PositiveIntegerField(default=0, null=True, blank=True)

    objects = DiscussionManager()

    class Meta:
        ordering = ["-created_at"]
        verbose_name = _("discussion")
        verbose_name_plural = _("discussions")
        permissions = (("can_view", "Can view"),)
        app_label = "discussions"
        abstract = True

    def __str__(self):
        return self.subject

    def update_counters(self, commit=True):
        self.recipients_count = self.recipients.count()
        self.messages_count = self.messages.count()

        if commit:
            self.save(update_fields=("recipients_count", "messages_count"))

    def save_recipients(self, to_user_list):
        from . import Recipient

        """
        Save the recipients for this message

        :param to_user_list:
            A list which elements are :class:`User` to whom the message is for.

        :return:
            Boolean indicating if any users are saved.

        """
        created = False
        for user in to_user_list:
            Recipient.objects.create(user=user, discussion=self)
            created = True
        return created

    def is_recipient(self, user):
        return user.pk in self.recipients.values_list("id", flat=True)

    def mark_as_read(self, user=None):
        from discussions.models import Recipient

        recipients = self.recipients.all()

        recipients = Recipient.objects.filter(discussion_id=self.pk)

        if user:
            recipients = recipients.filter(user_id=user.pk)

        recipients.update(read_at=tznow(), status=Recipient.STATUS.read)

    def add_message(self, body, sender=None, commit=True):
        from . import Message

        if not sender:
            sender = self.sender

        m = Message(sender=sender, body=body, discussion=self)
        m.save()

        if not self.is_recipient(sender):
            self.save_recipients([sender])

        self.recipient_set.exclude(user=sender).update(status=Recipient.STATUS.unread)

        self.updated_at = tznow()

        if commit:
            self.save()

        return m

    def get_absolute_url(self):
        return reverse("discussions_detail", kwargs={"discussion_id": self.pk})

    def can_view(self, user):
        if not user or not user.is_authenticated:
            return False

        if user.is_staff or user.is_superuser or self.is_recipient(user):
            return True

        return user.has_perm("discussions.can_view")

    def delete_recipient(self, user):
        if user.pk == self.sender_id:
            return False

        self.recipients.through.objects.filter(
            discussion=self.pk, user=user.pk
        ).delete()

        return True


@python_2_unicode_compatible
class Message(models.Model):
    """ Private message model, from user to user(s) """

    sender = models.ForeignKey(
        AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="sent_messages",
        verbose_name=_("sender"),
    )

    discussion = models.ForeignKey(
        get_model_string("Discussion"),
        on_delete=models.PROTECT,
        related_name="messages",
        verbose_name=_("discussion"),
    )

    body = models.TextField(_("body"))

    sent_at = models.DateTimeField(_("sent at"), auto_now_add=True)

    sender_deleted_at = models.DateTimeField(
        _("sender deleted at"), null=True, blank=True
    )

    objects = MessageManager()

    class Meta:
        ordering = ["-sent_at"]
        verbose_name = _("message")
        verbose_name_plural = _("messages")
        app_label = "discussions"
        abstract = True

    def __str__(self):
        """ Human representation, displaying first ten words of the body. """
        from ..compat import truncate_words

        truncated_body = truncate_words(self.body, 10)
        return "%(truncated_body)s" % {"truncated_body": truncated_body}


@python_2_unicode_compatible
class Folder(models.Model):
    name = models.CharField(max_length=255, verbose_name=_("Name"))

    created_at = models.DateTimeField(_("created at"), auto_now_add=True)

    user = models.ForeignKey(
        AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="discussion_folders"
    )

    class Meta:
        ordering = ["-created_at"]
        verbose_name = _("folder")
        verbose_name_plural = _("folders")
        app_label = "discussions"
        abstract = True

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse("discussions_folder_detail", kwargs={"folder_id": self.pk})



from __future__ import unicode_literals

from django.views.generic import DetailView, FormView, ListView
from django.views.generic.base import View
from django.views.generic.edit import FormMixin, CreateView, UpdateView, DeleteView
from django.urls import reverse
from django.shortcuts import get_object_or_404, redirect
from django.contrib import messages
from django.utils.translation import ugettext_lazy as _, ungettext
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.http import Http404
from django.utils.functional import cached_property
from django.utils.timezone import now as tznow

from ..models import Discussion, Folder, Recipient
from ..forms import ComposeForm, ReplyForm, FolderForm
from ..helpers import lookup_discussions, lookup_profiles
from .. import settings
from ..compat import User

from pure_pagination.paginator import Paginator


class DiscussionListView(ListView):
    template_name = "discussions/list.html"
    paginate_by = settings.PAGINATE_BY
    model = Recipient
    context_object_name = "recipient_list"
    paginator_class = Paginator

    @cached_property
    def user(self):
        return self.request.user

    @cached_property
    def folder(self):
        folder = None

        folder_id = self.kwargs.get("folder_id")

        if folder_id:
            folder = get_object_or_404(
                Folder.objects.filter(user=self.user), pk=folder_id
            )

        return folder

    def get_base_queryset(self):
        qs = (
            self.model.objects.filter(user=self.user)
            .order_by("-discussion__updated_at", "-discussion__created_at")
            .select_related("user")
        )

        return qs

    def get_queryset(self):
        if self.folder:
            qs = (
                self.get_base_queryset()
                .filter(folder=self.folder)
                .exclude(status=self.model.STATUS.deleted)
            )
        else:
            qs = (
                self.get_base_queryset()
                .exclude(status=self.model.STATUS.deleted)
                .exclude(folder__isnull=False)
            )

        return qs

    def get_context_data(self, **kwargs):
        context = super(DiscussionListView, self).get_context_data(**kwargs)

        lookup_discussions(context[self.context_object_name])
        lookup_profiles(context[self.context_object_name])

        context["folder"] = self.folder

        return context


class FoldersListView(ListView):
    template_name = "discussions/folder/list.html"
    paginate_by = settings.PAGINATE_BY
    model = Folder
    context_object_name = "folder_list"
    paginator_class = Paginator

    def get_queryset(self):
        qs = self.model.objects.filter(user=self.request.user).select_related("user")

        return qs


class DiscussionDetailView(DetailView, FormMixin):
    pk_url_kwarg = "discussion_id"
    model = Discussion
    context_object_name = "discussion"
    template_name = "discussions/detail.html"
    form_class = ReplyForm

    def get_success_url(self):
        return self.object.get_absolute_url()

    def is_allowed(self, user):
        return self.object.can_view(user)

    def get_context_data(self, **kwargs):
        data = super(DiscussionDetailView, self).get_context_data(**kwargs)

        if not self.is_allowed(self.request.user):
            raise Http404

        self.object.mark_as_read(self.request.user)

        recipients = self.object.recipients.all()

        return dict(
            data,
            **{
                "recipient_list": recipients,
                "form": self.get_form(self.get_form_class()),
                "message_list": self.get_messages(self.object),
            }
        )

    def get_messages(self, discussion):
        return discussion.messages.order_by("sent_at")

    def get_template_names(self):
        return [self.template_name]

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()

        form_class = self.get_form_class()
        form = self.get_form(form_class)

        if form.is_valid():
            return self.form_valid(form)

        return self.form_invalid(form)

    def form_valid(self, form):
        form.save(self.request.user)

        return super(DiscussionDetailView, self).form_valid(form)

    def get_form_kwargs(self):
        return dict(
            super(DiscussionDetailView, self).get_form_kwargs(),
            **{"discussion": self.object}
        )


class MessageComposeView(FormView):
    form_class = ComposeForm
    template_name = "discussions/form.html"
    submit_key = "submit"

    def get_template_names(self):
        return [self.template_name]

    @cached_property
    def recipients(self):
        recipients = None

        if self.request.method == "POST":
            recipients = self.request.POST.get("recipients", None)
            if recipients is None and self.kwargs.get("recipients"):
                recipients = self.kwargs.get("recipients")
        elif self.kwargs.get("recipients"):
            recipients = self.kwargs.get("recipients")

        if recipients:
            username_list = [r.strip() for r in recipients.split("+")]
            recipients = [u for u in User.objects.filter(username__in=username_list)]

        return recipients

    def get_initial(self):
        initial = {"to": self.recipients}

        return initial.copy()

    def get_context_data(self, **kwargs):
        data = super(MessageComposeView, self).get_context_data(**kwargs)

        return dict(
            data,
            **{"recipients": self.recipients if hasattr(self, "recipients") else None}
        )

    def form_valid(self, form):
        self.object = form.save(self.request.user)

        return redirect(self.get_success_url(form))

    def get_success_url(self, form):
        requested_redirect = self.request.POST.get(REDIRECT_FIELD_NAME, False)

        # Redirect mechanism
        redirect_to = reverse("discussions_list")
        if requested_redirect:
            redirect_to = requested_redirect
        elif self.kwargs.get("success_url"):
            redirect_to = self.kwargs.get("success_url")
        elif len(form.cleaned_data["to"]) == 1:
            redirect_to = reverse(
                "discussions_detail", kwargs={"discussion_id": self.object.pk}
            )

        return redirect_to

    def post(self, request, *args, **kwargs):
        if self.submit_key in request.POST:
            return super(MessageComposeView, self).post(request, *args, **kwargs)

        return self.get(request, *args, **kwargs)


class DiscussionBulkMixin(object):
    def valid_ids(self, ids):
        valid_discussion_id_list = set()
        for pk in ids:
            try:
                valid_pk = int(pk)
            except (TypeError, ValueError):
                pass
            else:
                valid_discussion_id_list.add(valid_pk)

        return valid_discussion_id_list


class DiscussionMoveView(DetailView, DiscussionBulkMixin):
    http_method_names = ["post"]
    model = Folder
    pk_url_kwarg = "folder_id"

    def get_queryset(self):
        return self.model.objects.filter(user=self.request.user)

    def post(self, *args, **kwargs):
        """
        A ``POST`` to move discussions into a folder.

        POST can have the following keys:

            ``discussions_ids``
                List of discussion id's that should be moved.
        """

        self.object = None

        if self.kwargs.get("folder_id"):
            self.object = self.get_object()
            self.get_context_data(object=self.object)

        discussion_ids = self.request.POST.getlist("discussion_ids")

        if discussion_ids:
            Recipient.objects.filter(
                discussion__in=self.valid_ids(discussion_ids), user=self.request.user
            ).update(folder=self.object)

        return redirect(reverse("discussions_list"))


class DiscussionMarkAsReadView(View, DiscussionBulkMixin):
    http_method_names = ["post"]

    def post(self, *args, **kwargs):
        """
        A ``POST`` to mark as read discussions.

        :param unread:
            A Boolean that if ``True`` unread discussions.

        POST can have the following keys:

            ``discussions_ids``
                List of discussion id's that should be marked as read.
        """

        discussion_ids = self.request.POST.getlist("discussion_ids")

        if discussion_ids:
            recipients = Recipient.objects.filter(
                discussion__in=self.valid_ids(discussion_ids), user=self.request.user
            ).exclude(status=Recipient.STATUS.read)
            for recipient in recipients:
                recipient.mark_as_read()

        folder_id = self.kwargs.get("folder_id")

        if folder_id:
            return redirect(
                reverse("discussions_list", kwargs={"folder_id": folder_id})
            )

        return redirect(reverse("discussions_list"))


class DiscussionMarkAsUnreadView(View, DiscussionBulkMixin):
    http_method_names = ["post"]

    def post(self, *args, **kwargs):
        """
        A ``POST`` to mark as unread discussions.

        POST can have the following keys:

            ``discussions_ids``
                List of discussion id's that should be marked as unread.
        """

        discussion_ids = self.request.POST.getlist("discussion_ids")

        if discussion_ids:
            recipients = Recipient.objects.filter(
                discussion__in=self.valid_ids(discussion_ids), user=self.request.user
            ).exclude(status=Recipient.STATUS.unread)
            for recipient in recipients:
                recipient.mark_as_unread()

        folder_id = self.kwargs.get("folder_id")

        if folder_id:
            return redirect(
                reverse("discussions_list", kwargs={"folder_id": folder_id})
            )

        return redirect(reverse("discussions_list"))


class DiscussionLeaveView(View, DiscussionBulkMixin):
    http_method_names = ["post"]

    def post(self, *args, **kwargs):
        """
        A ``POST`` to leave discussions.

        POST can have the following keys:

            ``discussions_ids``
                List of discussion id's that should be deleted.

            ``next``
                String containing the URI which to redirect to after the keys are
                removed. Redirect defaults to the inbox view.

        The ``next`` value can also be supplied in the URI with ``?next=<value>``.

        """

        discussion_ids = self.request.POST.getlist("discussion_ids")
        redirect_to = self.request.POST.get("next", False)

        if discussion_ids:
            for pk in self.valid_ids(discussion_ids):
                discussion = get_object_or_404(Discussion, pk=pk)

                result = discussion.delete_recipient(self.request.user)

                if result:
                    messages.success(
                        self.request,
                        _('You have successfully left the discussion "%s"')
                        % discussion,
                        fail_silently=True,
                    )
                else:
                    messages.error(
                        self.request,
                        _('You have created the discussion "%s", you cannot leave it')
                        % discussion,
                        fail_silently=True,
                    )

        if redirect_to:
            return redirect(redirect_to)

        return redirect(reverse("discussions_list"))


class DiscussionRemoveView(View, DiscussionBulkMixin):
    http_method_names = ["post"]

    def post(self, *args, **kwargs):
        """
        A ``POST`` to remove discussions.

        :param undo:
            A Boolean that if ``True`` unremoves discussions.

        POST can have the following keys:

            ``discussions_ids``
                List of discussion id's that should be deleted.

            ``next``
                String containing the URI which to redirect to after the keys are
                removed. Redirect defaults to the inbox view.

        The ``next`` value can also be supplied in the URI with ``?next=<value>``.

        """
        discussion_ids = self.request.POST.getlist("discussion_ids")
        redirect_to = self.request.POST.get("next", False)
        undo = self.kwargs.get("undo", False)

        if discussion_ids:
            # Delete all the messages, if they belong to the user.
            now = tznow()
            changed_message_list = set()
            for pk in self.valid_ids(discussion_ids):
                discussion = get_object_or_404(Discussion, pk=pk)

                # Check if the user is the owner
                if discussion.sender_id == self.request.user.pk:
                    if undo:
                        discussion.sender_deleted_at = None
                    else:
                        discussion.sender_deleted_at = now
                    discussion.save()
                    changed_message_list.add(discussion.pk)

                # Check if the user is a recipient of the message
                recipients = discussion.recipient_set.filter(
                    user=self.request.user, discussion=discussion
                )

                if recipients:
                    for recipient in recipients:
                        if undo:
                            recipient.mark_as_read()
                        else:
                            recipient.mark_as_deleted()

                    changed_message_list.add(discussion.pk)

            # Send messages
            if len(changed_message_list) > 0:
                if undo:
                    message = ungettext(
                        "This discussion has been succesfully restored",
                        "These discussions have been succesfully restored",
                        len(changed_message_list),
                    )
                else:
                    message = ungettext(
                        "This discussion has been successfully removed",
                        "These discussions have been successfully removed",
                        len(changed_message_list),
                    )

                messages.success(self.request, message, fail_silently=True)

        if redirect_to:
            return redirect(redirect_to)

        folder_id = self.kwargs.get("folder_id")

        if folder_id:
            return redirect(
                reverse("discussions_list", kwargs={"folder_id": folder_id})
            )

        return redirect(reverse("discussions_list"))


class FolderCreateView(CreateView):
    form_class = FolderForm
    template_name = "discussions/folder/create.html"

    def form_valid(self, form):
        self.object = form.save(self.request.user)

        return redirect(self.get_success_url())

    def get_success_url(self):
        message = _("Your folder has been successfully created")
        messages.success(self.request, message, fail_silently=True)
        return reverse("discussions_list")


class FolderUpdateView(UpdateView):
    form_class = FolderForm
    model = Folder
    template_name = "discussions/folder/update.html"
    pk_url_kwarg = "folder_id"

    def get_queryset(self):
        return self.model.objects.filter(
            pk=self.kwargs.get(self.pk_url_kwarg), user=self.request.user
        )

    def form_valid(self, form):
        self.object = form.save(self.request.user)

        return redirect(self.get_success_url())

    def get_success_url(self):
        message = _("You have successfully updated this folder")
        messages.success(self.request, message, fail_silently=True)
        return reverse("discussions_list", kwargs={"folder_id": self.object.pk})


class FolderRemoveView(DeleteView):
    http_method_names = ["post"]
    model = Folder
    pk_url_kwarg = "folder_id"

    def get_success_url(self):
        return reverse("discussions_list")

    def get_queryset(self):
        return self.model.objects.filter(user=self.request.user)



import django

from django.contrib.auth import get_user_model


# Django 1.6+ compatibility
if django.VERSION >= (1, 4):
    from django.utils.text import Truncator

    def truncate_words(s, num, end_text="..."):
        truncate = end_text and " %s" % end_text or ""
        return Truncator(s).words(num, truncate=truncate)


else:
    from django.utils.text import truncate_words

User = get_user_model()

__all__ = ["User", "truncate_words"]



from django import forms
from django.forms import widgets
from django.utils.translation import ugettext_lazy as _

from .compat import User


class CommaSeparatedUserInput(widgets.Input):
    input_type = "text"

    def render(self, name, value, attrs=None):
        if value is None:
            value = ""
        elif isinstance(value, (list, tuple)):
            value = ", ".join([user.username for user in value])
        return super(CommaSeparatedUserInput, self).render(name, value, attrs)


class CommaSeparatedUserField(forms.Field):
    """
    A :class:`CharField` that exists of comma separated usernames.

    :param recipient_filter:
        Optional function which receives as :class:`User` as parameter. The
        function should return ``True`` if the user is allowed or ``False`` if
        the user is not allowed.

    :return:
        A list of :class:`User`.

    """

    widget = CommaSeparatedUserInput

    def __init__(self, *args, **kwargs):
        recipient_filter = kwargs.pop("recipient_filter", None)
        self._recipient_filter = recipient_filter
        super(CommaSeparatedUserField, self).__init__(*args, **kwargs)

    def clean(self, value):
        super(CommaSeparatedUserField, self).clean(value)

        names = set(value.split(","))
        names_set = set([name.strip() for name in names])
        users = list(User.objects.filter(username__in=names_set))

        # Check for unknown names.
        unknown_names = names_set ^ set([user.username for user in users])

        recipient_filter = self._recipient_filter
        invalid_users = []
        if recipient_filter is not None:
            for r in users:
                if recipient_filter(r) is False:
                    users.remove(r)
                    invalid_users.append(r.username)

        if unknown_names or invalid_users:
            humanized_usernames = ", ".join(sorted(list(unknown_names) + invalid_users))

            raise forms.ValidationError(
                _("The following usernames are incorrect: %(users)s.")
                % {"users": humanized_usernames}
            )

        return users



from __future__ import unicode_literals

from django.test import TestCase
from django.template import Context, Template

from ..compat import User


class TemplateTagsTests(TestCase):
    fixtures = ["users.json", "messages.json"]

    def test_get_unread_message_count_for(self):
        user = User.objects.get(pk=1)

        t = Template(
            "{% load discussions_tags %}{% get_unread_message_count_for user as message_count %}{{ message_count }}"
        )
        c = Context({"user": user})
        self.assertEqual("1", t.render(c))

    def test_get_unread_message_count_between(self):
        thoas = User.objects.get(pk=1)
        ampelmann = User.objects.get(pk=2)

        t = Template(
            "{% load discussions_tags %}{% get_unread_message_count_between user_1 and user_2 as message_count %}{{ message_count }}"
        )
        c = Context({"user_1": thoas, "user_2": ampelmann})
        self.assertEqual("0", t.render(c))

    def test_get_folders_for(self):
        user = User.objects.get(pk=1)

        t = Template(
            "{% load discussions_tags %}{% get_folders_for user as folders %}{% for folder in folders %}{{ folder.name }}{% endfor %}"
        )
        c = Context({"user": user})

        self.assertEqual("My folder", t.render(c))

        c = Context({"user": User.objects.get(pk=2)})

        self.assertEqual("", t.render(c))  # return nothing



from ..utils import load_class
from .. import settings


class DiscussionSentView(load_class(settings.DISCUSSION_LIST_VIEW)):
    template_name = "discussions/sent.html"

    def get_queryset(self):
        return (
            self.get_base_queryset()
            .filter(discussion__sender=self.user, folder=self.folder)
            .exclude(status=self.model.STATUS.deleted)
        )


class DiscussionUnreadView(load_class(settings.DISCUSSION_LIST_VIEW)):
    template_name = "discussions/unread.html"

    def get_queryset(self):
        return self.get_base_queryset().filter(
            status=self.model.STATUS.unread, folder=self.folder
        )


class DiscussionReadView(load_class(settings.DISCUSSION_LIST_VIEW)):
    template_name = "discussions/read.html"

    def get_queryset(self):
        return self.get_base_queryset().filter(
            status=self.model.STATUS.read, folder=self.folder
        )


class DiscussionDeletedView(load_class(settings.DISCUSSION_LIST_VIEW)):
    template_name = "discussions/deleted.html"

    def get_queryset(self):
        return self.get_base_queryset().filter(
            status=self.model.STATUS.deleted, folder=self.folder
        )



from . import base

from discussions.managers import RecipientManager


class Recipient(base.Recipient):
    class Meta(base.Recipient.Meta):
        abstract = False

    objects = RecipientManager()



from discussions.utils import load_class

from discussions import settings

Recipient = load_class(settings.RECIPIENT_MODEL)

Message = load_class(settings.MESSAGE_MODEL)

Folder = load_class(settings.FOLDER_MODEL)

Discussion = load_class(settings.DISCUSSION_MODEL)



from django.test import TestCase
from django.urls import reverse

from ..forms import ComposeForm, FolderForm
from ..models import Message, Recipient, Discussion, Folder
from ..compat import User


class DiscussionsViewsTests(TestCase):
    fixtures = ["users.json", "messages.json"]

    def _test_login(self, named_url, **kwargs):
        """ Test that the view requires login """
        response = self.client.get(reverse(named_url, **kwargs))
        self.assertEqual(response.status_code, 302)

    def test_compose(self):
        """ A ``GET`` to the compose view """
        # Login is required.
        self._test_login("discussions_compose")

        # Sign in
        self.client.login(username="thoas", password="$ecret")
        response = self.client.get(reverse("discussions_compose"))

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "discussions/form.html")

        self.assertTrue(isinstance(response.context["form"], ComposeForm))

    def test_compose_post(self):
        """ ``POST`` to the compose view """
        self.client.login(username="thoas", password="$ecret")

        valid_data = {"to": "ampelmann", "body": "Hi mister", "subject": "Hi"}

        response = self.client.post(reverse("discussions_compose"), data=valid_data)

        self.assertEqual(response.status_code, 200)

        valid_data["submit"] = "submit"

        response = self.client.post(reverse("discussions_compose"), data=valid_data)

        self.assertRedirects(
            response, reverse("discussions_detail", kwargs={"discussion_id": 3})
        )

        # Check for a requested redirect
        valid_data["next"] = reverse("discussions_compose")
        response = self.client.post(reverse("discussions_compose"), data=valid_data)
        self.assertRedirects(response, valid_data["next"])

    def test_compose_recipients_get(self):
        """ A ``GET`` to the compose view with recipients """
        self.client.login(username="thoas", password="$ecret")

        valid_recipients = "thoas+oleiade"

        # Test valid recipients
        response = self.client.get(
            reverse("discussions_compose", kwargs={"recipients": valid_recipients})
        )

        self.assertEqual(response.status_code, 200)

        # Test the users
        oleiade = User.objects.get(username="oleiade")
        thoas = User.objects.get(username="thoas")
        self.assertEqual(response.context["recipients"][0], oleiade)
        self.assertEqual(response.context["recipients"][1], thoas)

        # Test that the initial data of the form is set.
        self.assertEqual(response.context["form"].initial["to"], [oleiade, thoas])

    def test_compose_recipients_post(self):
        """ A ``POST`` to the compose view with recipients """
        self.client.login(username="thoas", password="$ecret")

        valid_recipients = "thoas+oleiade"

        # Test valid recipients
        response = self.client.post(
            reverse("discussions_compose"), data={"recipients": valid_recipients}
        )

        self.assertEqual(response.status_code, 200)

        # Test the users
        oleiade = User.objects.get(username="oleiade")
        thoas = User.objects.get(username="thoas")
        self.assertEqual(response.context["recipients"][0], oleiade)
        self.assertEqual(response.context["recipients"][1], thoas)

        # Test that the initial data of the form is set.
        self.assertEqual(response.context["form"].initial["to"], [oleiade, thoas])

    def test_discussion_detail(self):
        """ A ``GET`` to the detail view """
        self._test_login("discussions_detail", kwargs={"discussion_id": 2})

        # Sign in
        self.client.login(username="ampelmann", password="$ecret")
        response = self.client.get(
            reverse("discussions_detail", kwargs={"discussion_id": 1})
        )

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "discussions/detail.html")

        # Test that the message is read.
        ampelmann = User.objects.get(pk=2)
        mr = Recipient.objects.get(
            discussion=Discussion.objects.get(pk=1), user=ampelmann
        )

        assert mr.read_at is not None

    def test_discussion_detail_new_message(self):
        self.client.login(username="ampelmann", password="$ecret")
        discussion = Discussion.objects.get(pk=1)

        response = self.client.post(
            discussion.get_absolute_url(), data={"body": "My reply"}
        )
        self.assertRedirects(response, discussion.get_absolute_url())

        self.assertEqual(discussion.messages.count(), 2)

        ampelmann = User.objects.get(username="ampelmann")

        message = discussion.messages.get(sender=ampelmann)

        self.assertEqual(message.body, "My reply")

    def test_valid_discussion_move(self):
        """ ``POST`` to move a discussion into a folder """
        # Test that sign in is required

        folder = Folder.objects.get(pk=1)

        url = reverse("discussions_move", kwargs={"folder_id": folder.pk})
        response = self.client.post(url)
        self.assertEqual(response.status_code, 302)

        # Sign in
        self.client.login(username="thoas", password="$ecret")

        # Test that only posts are allowed
        response = self.client.get(url)
        self.assertEqual(response.status_code, 405)

        response = self.client.post(url, data={"discussion_ids": "1"})
        self.assertRedirects(response, reverse("discussions_list"))

        user = User.objects.get(username="thoas")

        recipient = Recipient.objects.get(discussion=1, user=user)

        self.assertEqual(recipient.folder, folder)

    def test_invalid_discussion_move(self):
        folder = Folder.objects.get(pk=1)

        url = reverse("discussions_move", kwargs={"folder_id": folder.pk})

        self.client.login(username="ampelmann", password="$ecret")

        response = self.client.post(url, data={"discussion_ids": "1"})

        self.assertEqual(response.status_code, 404)

    def test_valid_discussion_mark_as_read(self):
        """ ``POST`` to mark a discussion as read """
        # Test that sign in is required
        response = self.client.post(reverse("discussions_mark_as_read"))
        self.assertEqual(response.status_code, 302)

        # Sign in
        self.client.login(username="thoas", password="$ecret")

        # Test that only posts are allowed
        response = self.client.get(reverse("discussions_mark_as_read"))
        self.assertEqual(response.status_code, 405)

        response = self.client.post(
            reverse("discussions_mark_as_read"), data={"discussion_ids": "1"}
        )
        self.assertRedirects(response, reverse("discussions_list"))

        recipient = Discussion.objects.get(pk=1).recipient_set.get(
            user=User.objects.get(pk=1)
        )

        self.assertTrue(recipient.is_read())

    def test_valid_discussion_mark_as_unread(self):
        """ ``POST`` to mark a discussion as unread"""
        # Test that sign in is required
        response = self.client.post(reverse("discussions_mark_as_unread"))
        self.assertEqual(response.status_code, 302)

        # Sign in
        self.client.login(username="thoas", password="$ecret")

        # Test that only posts are allowed
        response = self.client.get(reverse("discussions_mark_as_unread"))
        self.assertEqual(response.status_code, 405)

        response = self.client.post(
            reverse("discussions_mark_as_unread"), data={"discussion_ids": "1"}
        )
        self.assertRedirects(response, reverse("discussions_list"))

        recipient = Discussion.objects.get(pk=1).recipient_set.get(
            user=User.objects.get(pk=1)
        )

        self.assertFalse(recipient.is_read())

    def test_valid_discussion_leave(self):
        """ ``POST`` to leave a discussion
        # Test that sign in is required'"""
        response = self.client.post(reverse("discussions_leave"))
        self.assertEqual(response.status_code, 302)

        # Sign in
        self.client.login(username="ampelmann", password="$ecret")

        # Test that only posts are allowed
        response = self.client.get(reverse("discussions_leave"))
        self.assertEqual(response.status_code, 405)

        response = self.client.post(
            reverse("discussions_leave"), data={"discussion_ids": [1]}
        )

        self.assertRedirects(response, reverse("discussions_list"))

        self.assertFalse(Discussion.objects.get(pk=1).recipients.filter(pk=2).exists())

    def test_invalid_discussion_leave_when_sender(self):
        """ ``POST`` to leave a discussion
        # Test that sign in is required'"""
        response = self.client.post(reverse("discussions_leave"))
        self.assertEqual(response.status_code, 302)

        # Sign in
        self.client.login(username="thoas", password="$ecret")

        # Test that only posts are allowed
        response = self.client.get(reverse("discussions_leave"))
        self.assertEqual(response.status_code, 405)

        response = self.client.post(
            reverse("discussions_leave"), data={"discussion_ids": [1]}
        )

        self.assertRedirects(response, reverse("discussions_list"))

        self.assertTrue(Discussion.objects.get(pk=1).recipients.filter(pk=2).exists())

    def test_valid_discussion_remove(self):
        """ ``POST`` to remove a discussion """
        # Test that sign in is required
        response = self.client.post(reverse("discussions_remove"))
        self.assertEqual(response.status_code, 302)

        # Sign in
        self.client.login(username="thoas", password="$ecret")

        # Test that only posts are allowed
        response = self.client.get(reverse("discussions_remove"))
        self.assertEqual(response.status_code, 405)

        response = self.client.post(
            reverse("discussions_remove"), data={"discussion_ids": "1"}
        )
        self.assertRedirects(response, reverse("discussions_list"))
        d = Discussion.objects.get(pk=1)
        assert d.sender_deleted_at is not None

        self.client.login(username="ampelmann", password="$ecret")
        response = self.client.post(
            reverse("discussions_remove"),
            data={"discussion_ids": "1", "next": reverse("discussions_list")},
        )
        self.assertRedirects(response, reverse("discussions_list"))
        ampelmann = User.objects.get(username="ampelmann")
        dr = d.recipient_set.get(user=ampelmann, discussion=d)

        assert dr.deleted_at is not None

    def test_invalid_discussion_remove(self):
        """ ``POST`` to remove an invalid discussion """
        # Sign in
        self.client.login(username="thoas", password="$ecret")

        bef_len = Message.objects.filter(sender_deleted_at__isnull=False).count()
        response = self.client.post(
            reverse("discussions_remove"), data={"discussion_ids": ["a", "b"]}
        )

        # The program should play nice, nothing happened.
        af_len = Message.objects.filter(sender_deleted_at__isnull=False).count()
        self.assertRedirects(response, reverse("discussions_list"))
        self.assertEqual(bef_len, af_len)

    def test_valid_discussion_remove_multiple(self):
        """ ``POST`` to remove multiple messages """
        # Sign in
        self.client.login(username="thoas", password="$ecret")
        response = self.client.post(
            reverse("discussions_remove"), data={"discussion_ids": [1, 2]}
        )
        self.assertRedirects(response, reverse("discussions_list"))

        discussion_list = Discussion.objects.filter(
            pk__in=["1", "2"], sender_deleted_at__isnull=False
        )

        # thoas has created only one discussion (see fixtures)
        self.assertEqual(discussion_list.count(), 1)

    def test_discussion_unremove(self):
        """ Unremove a discussion """
        self.client.login(username="thoas", password="$ecret")

        # Delete a discussion as owner
        response = self.client.post(
            reverse("discussions_unremove"), data={"discussion_pks": [1]}
        )

        self.assertRedirects(response, reverse("discussions_list"))

        # Delete the message as a recipient
        response = self.client.post(
            reverse("discussions_unremove"), data={"discussion_pks": [2]}
        )

        self.assertRedirects(response, reverse("discussions_list"))

    def test_discussion_list(self):
        """ ``GET`` the discussion list for a user """
        self._test_login("discussions_list")

        self.client.login(username="thoas", password="$ecret")
        response = self.client.get(reverse("discussions_list"))
        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "discussions/list.html")

    def test_discussion_sent(self):
        """ ``GET`` the message list for a user """
        self._test_login("discussions_sent")

        self.client.login(username="thoas", password="$ecret")
        response = self.client.get(reverse("discussions_sent"))
        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "discussions/sent.html")

    def test_discussion_deleted(self):
        """ ``GET`` the message list for a user """
        self._test_login("discussions_deleted")

        self.client.login(username="thoas", password="$ecret")
        response = self.client.get(reverse("discussions_deleted"))
        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "discussions/deleted.html")

    def test_discussion_unread(self):
        """ ``GET`` the message list for a user """
        self._test_login("discussions_unread")

        self.client.login(username="thoas", password="$ecret")
        response = self.client.get(reverse("discussions_unread"))
        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "discussions/unread.html")

    def test_discussion_read(self):
        """ ``GET`` the message list for a user """
        self._test_login("discussions_unread")

        self.client.login(username="thoas", password="$ecret")
        response = self.client.get(reverse("discussions_read"))
        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "discussions/read.html")

    def test_discussion_detail_between_two_users(self):
        """ ``GET`` to a detail page between two users """
        self._test_login("discussions_list")
        self.client.login(username="thoas", password="$ecret")

        response = self.client.get(reverse("discussions_list"))

        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "discussions/list.html")

        # Check that all the messages are marked as read.
        thoas = User.objects.get(pk=1)
        oleiade = User.objects.get(pk=3)
        unread_messages = Recipient.objects.filter(
            user=thoas, discussion__sender=oleiade, read_at__isnull=True
        )

        self.assertEqual(len(unread_messages), 0)

    def test_folder_list(self):
        """ ``GET`` the discussion list for a user """
        self._test_login("discussions_folders_list")

        self.client.login(username="thoas", password="$ecret")
        response = self.client.get(reverse("discussions_folders_list"))
        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "discussions/folder/list.html")

    def test_folder_create(self):
        self._test_login("discussions_folder_create")

        self.client.login(username="thoas", password="$ecret")

        response = self.client.get(reverse("discussions_folder_create"))

        self.assertEqual(response.status_code, 200)

        self.assertTrue(isinstance(response.context["form"], FolderForm))
        self.assertTemplateUsed(response, "discussions/folder/create.html")

        data = {"name": "My folder"}

        response = self.client.post(reverse("discussions_folder_create"), data=data)

        self.assertEqual(Folder.objects.count(), 2)

        folder = Folder.objects.get(pk=2)

        self.assertEqual(folder.name, data["name"])

        self.assertRedirects(response, reverse("discussions_list"))

    def test_folder_update(self):
        folder = Folder.objects.get(pk=1)

        self._test_login("discussions_folder_update", kwargs={"folder_id": folder.pk})

        self.client.login(username="thoas", password="$ecret")

        response = self.client.get(
            reverse("discussions_folder_update", kwargs={"folder_id": folder.pk})
        )

        self.assertEqual(response.status_code, 200)

        self.assertTrue(isinstance(response.context["form"], FolderForm))
        self.assertTemplateUsed(response, "discussions/folder/update.html")

        data = {"name": "My folder new name"}

        response = self.client.post(
            reverse("discussions_folder_update", kwargs={"folder_id": folder.pk}),
            data=data,
        )

        folder = Folder.objects.get(pk=1)

        self.assertEqual(folder.name, data["name"])

        self.assertRedirects(
            response, reverse("discussions_list", kwargs={"folder_id": folder.pk})
        )

    def test_discussion_move_to_folder(self):
        response = self.client.post(
            reverse("discussions_move", kwargs={"folder_id": 1}),
            data={"discussion_ids": [2]},
        )
        self.assertEqual(response.status_code, 302)

        # Sign in
        self.client.login(username="thoas", password="$ecret")

        # Test that only posts are allowed
        response = self.client.get(
            reverse("discussions_move", kwargs={"folder_id": 1}),
            data={"discussion_ids": [2]},
        )
        self.assertEqual(response.status_code, 405)

        response = self.client.post(
            reverse("discussions_move", kwargs={"folder_id": 1}),
            data={"discussion_ids": [2]},
        )
        recipient = Recipient.objects.get(pk=2)

        folder = Folder.objects.get(pk=1)

        self.assertEqual(recipient.folder, folder)

    def test_discussion_move_to_mailbox(self):
        response = self.client.post(
            reverse("discussions_move"), data={"discussion_ids": [3]}
        )
        self.assertEqual(response.status_code, 302)

        # Sign in
        self.client.login(username="thoas", password="$ecret")

        # Test that only posts are allowed
        response = self.client.get(
            reverse("discussions_move"), data={"discussion_ids": [3]}
        )
        self.assertEqual(response.status_code, 405)

        response = self.client.post(
            reverse("discussions_move"), data={"discussion_ids": [3]}
        )
        recipient = Recipient.objects.get(pk=3)

        self.assertEqual(recipient.folder, None)

    def test_folder_remove_pk_does_not_exist(self):
        # Sign in
        self.client.login(username="thoas", password="$ecret")

        response = self.client.post(
            reverse("discussions_folder_remove", kwargs={"folder_id": 35})
        )

        self.assertEqual(response.status_code, 404)

    def test_valid_folder_remove(self):
        response = self.client.post(
            reverse("discussions_folder_remove", kwargs={"folder_id": 1})
        )
        self.assertEqual(response.status_code, 302)

        # Sign in
        self.client.login(username="thoas", password="$ecret")

        # Test that only posts are allowed
        response = self.client.get(
            reverse("discussions_folder_remove", kwargs={"folder_id": 1})
        )
        self.assertEqual(response.status_code, 405)

        folder = Folder.objects.get(pk=1)
        related_recipient = Recipient.objects.filter(folder=folder)

        response = self.client.post(
            reverse("discussions_folder_remove", kwargs={"folder_id": 1})
        )

        for recipient in related_recipient:
            self.assertEqual(recipient.folder, None)

        self.assertRedirects(response, reverse("discussions_list"))

        self.assertFalse(Folder.objects.filter(pk=1).exists())



from django.test import TestCase
from django import forms

from ..fields import CommaSeparatedUserField


class CommaSeparatedTestForm(forms.Form):
    users = CommaSeparatedUserField()

    def __init__(self, *args, **kwargs):
        super(CommaSeparatedTestForm, self).__init__(*args, **kwargs)
        self.fields["users"]._recipient_filter = self.filter_jane

    def filter_jane(self, user):
        if user.username == "ampelmann":
            return False
        return True


class CommaSeperatedFieldTests(TestCase):
    fixtures = ["users.json"]

    def test_invalid_data(self):
        # Test invalid data supplied to the field.
        invalid_data_dicts = [
            # Empty username
            {"data": {"users": ""}, "error": ("users", [u"This field is required."])},
            # No data
            {"data": {}, "error": ("users", [u"This field is required."])},
            # A list
            {"data": {"users": []}, "error": ("users", [u"This field is required."])},
            # Forbidden username
            {
                "data": {"users": "ampelmann"},
                "error": (
                    "users",
                    [u"The following usernames are incorrect: ampelmann."],
                ),
            },
            # Non-existant username
            {
                "data": {"users": "foo"},
                "error": ("users", [u"The following usernames are incorrect: foo."]),
            },
            # Multiple invalid usernames
            {
                "data": {"users": "foo, bar"},
                "error": (
                    "users",
                    [u"The following usernames are incorrect: bar, foo."],
                ),
            },
            # Valid and invalid
            {
                "data": {"users": "foo, thoas, bar"},
                "error": (
                    "users",
                    [u"The following usernames are incorrect: bar, foo."],
                ),
            },
            # Extra whitespace
            {
                "data": {"users": "foo,    thoas  "},
                "error": ("users", [u"The following usernames are incorrect: foo."]),
            },
        ]
        for invalid_dict in invalid_data_dicts:
            form = CommaSeparatedTestForm(data=invalid_dict["data"])
            assert form.is_valid() is False
            self.assertEqual(
                form.errors[invalid_dict["error"][0]], invalid_dict["error"][1]
            )



# -*- coding: utf-8 -*-
import six

from collections import defaultdict

from importlib import import_module

from django.core import exceptions
from django.conf import settings
from django.apps import apps

from . import settings as defaults


CLASS_PATH_ERROR = "django-discussions is unable to interpret settings value for %s. " "%s should be in the form of a tupple: " "('path.to.models.Class', 'app_label')."


def load_class(class_path, setting_name=None):
    """
    Loads a class given a class_path. The setting value may be a string or a
    tuple.

    The setting_name parameter is only there for pretty error output, and
    therefore is optional
    """
    if not isinstance(class_path, six.string_types):
        try:
            class_path, app_label = class_path
        except:
            if setting_name:
                raise exceptions.ImproperlyConfigured(
                    CLASS_PATH_ERROR % (setting_name, setting_name)
                )
            else:
                raise exceptions.ImproperlyConfigured(
                    CLASS_PATH_ERROR % ("this setting", "It")
                )

    try:
        class_module, class_name = class_path.rsplit(".", 1)
    except ValueError:
        if setting_name:
            txt = "%s isn't a valid module. Check your %s setting" % (
                class_path,
                setting_name,
            )
        else:
            txt = "%s isn't a valid module." % class_path
        raise exceptions.ImproperlyConfigured(txt)

    try:
        mod = import_module(class_module)
    except ImportError as e:
        if setting_name:
            txt = 'Error importing backend %s: "%s". Check your %s setting' % (
                class_module,
                e,
                setting_name,
            )
        else:
            txt = 'Error importing backend %s: "%s".' % (class_module, e)

        raise exceptions.ImproperlyConfigured(txt)

    try:
        clazz = getattr(mod, class_name)
    except AttributeError:
        if setting_name:
            txt = (
                'Backend module "%s" does not define a "%s" class. Check'
                " your %s setting" % (class_module, class_name, setting_name)
            )
        else:
            txt = 'Backend module "%s" does not define a "%s" class.' % (
                class_module,
                class_name,
            )
        raise exceptions.ImproperlyConfigured(txt)
    return clazz


def get_profile_model():
    """
    Return the model class for the currently-active user profile
    model, as defined by the ``AUTH_PROFILE_MODULE`` setting.

    :return: The model that is used as profile.

    """
    if (not hasattr(settings, "AUTH_PROFILE_MODULE")) or (
        not settings.AUTH_PROFILE_MODULE
    ):
        return None

    return apps.get_model(*settings.AUTH_PROFILE_MODULE.split("."))


def queryset_to_dict(qs, key="pk", singular=True):
    """
    Given a queryset will transform it into a dictionary based on ``key``.
    """
    if singular:
        result = {}
        for u in qs:
            result.setdefault(getattr(u, key), u)
    else:
        result = defaultdict(list)
        for u in qs:
            result[getattr(u, key)].append(u)
    return result


def get_model_string(model_name):
    """
    Returns the model string notation Django uses for lazily loaded ForeignKeys
    (eg 'auth.User') to prevent circular imports.

    This is needed to allow our crazy custom model usage.
    """
    setting_name = "DISCUSSIONS_%s_MODEL" % model_name.upper().replace("_", "")
    class_path = getattr(defaults, setting_name, None)

    if not class_path:
        return "discussions.%s" % model_name
    elif isinstance(class_path, basestring):
        parts = class_path.split(".")
        try:
            index = parts.index("models") - 1
        except ValueError:
            raise exceptions.ImproperlyConfigured(
                CLASS_PATH_ERROR % (setting_name, setting_name)
            )
        app_label, model_name = parts[index], parts[-1]
    else:
        try:
            class_path, app_label = class_path
            model_name = class_path.split(".")[-1]
        except:
            raise exceptions.ImproperlyConfigured(
                CLASS_PATH_ERROR % (setting_name, setting_name)
            )

    return "%s.%s" % (app_label, model_name)



from django.conf import settings


PAGINATE_BY = getattr(settings, "DISCUSSIONS_PAGINATE_BY", 20)

RECIPIENT_MODEL = getattr(
    settings, "DISCUSSIONS_RECIPIENT_MODEL", "discussions.models.recipient.Recipient"
)
DISCUSSION_MODEL = getattr(
    settings, "DISCUSSIONS_DISCUSSION_MODEL", "discussions.models.discussion.Discussion"
)
FOLDER_MODEL = getattr(
    settings, "DISCUSSIONS_FOLDER_MODEL", "discussions.models.folder.Folder"
)
MESSAGE_MODEL = getattr(
    settings, "DISCUSSIONS_MESSAGE_MODEL", "discussions.models.message.Message"
)

DISCUSSION_LIST_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_LIST_VIEW",
    "discussions.views.base.DiscussionListView",
)
DISCUSSION_SENT_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_SENT_VIEW",
    "discussions.views.status.DiscussionSentView",
)
DISCUSSION_DELETED_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_DELETED_VIEW",
    "discussions.views.status.DiscussionDeletedView",
)
DISCUSSION_READ_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_READ_VIEW",
    "discussions.views.status.DiscussionReadView",
)
DISCUSSION_UNREAD_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_UNREAD_VIEW",
    "discussions.views.status.DiscussionUnreadView",
)
DISCUSSION_DETAIL_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_DETAIL_VIEW",
    "discussions.views.base.DiscussionDetailView",
)
DISCUSSION_REMOVE_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_REMOVE_VIEW",
    "discussions.views.base.DiscussionRemoveView",
)
DISCUSSION_MARK_AS_READ_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_MARK_READ_AS_VIEW",
    "discussions.views.base.DiscussionMarkAsReadView",
)

DISCUSSION_MARK_AS_UNREAD_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_MARK_UNREAD_AS_VIEW",
    "discussions.views.base.DiscussionMarkAsUnreadView",
)

DISCUSSION_MOVE_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_MOVE_VIEW",
    "discussions.views.base.DiscussionMoveView",
)
MESSAGE_COMPOSE_VIEW = getattr(
    settings,
    "DISCUSSIONS_MESSAGE_COMPOSE_VIEW",
    "discussions.views.base.MessageComposeView",
)
FOLDER_LIST_VIEW = getattr(
    settings, "DISCUSSIONS_FOLDER_LIST_VIEW", "discussions.views.base.FoldersListView"
)
FOLDER_CREATE_VIEW = getattr(
    settings,
    "DISCUSSIONS_FOLDER_CREATE_VIEW",
    "discussions.views.base.FolderCreateView",
)
FOLDER_UPDATE_VIEW = getattr(
    settings,
    "DISCUSSIONS_FOLDER_UPDATE_VIEW",
    "discussions.views.base.FolderUpdateView",
)
FOLDER_REMOVE_VIEW = getattr(
    settings,
    "DISCUSSIONS_FOLDER_REMOVE_VIEW",
    "discussions.views.base.FolderRemoveView",
)
DISCUSSION_LEAVE_VIEW = getattr(
    settings,
    "DISCUSSIONS_DISCUSSION_LEAVE_VIEW",
    "discussions.views.base.DiscussionLeaveView",
)

COMPOSE_FORM = getattr(
    settings, "DISCUSSIONS_COMPOSE_FORM", "discussions.forms.base.ComposeForm"
)

REPLY_FORM = getattr(
    settings, "DISCUSSIONS_REPLY_FORM", "discussions.forms.base.ReplyForm"
)

FOLDER_FORM = getattr(
    settings, "DISCUSSIONS_FOLDER_FORM", "discussions.forms.base.FolderForm"
)

PRE_FILTER = getattr(
    settings, "DISCUSSIONS_PRE_FILTER", "django.contrib.auth.decorators.login_required"
)



from . import base

from discussions.managers import DiscussionManager


class Discussion(base.Discussion):
    class Meta(base.Discussion.Meta):
        abstract = False

    objects = DiscussionManager()



from django.db import models
from django.db.models import Q
from django.db.models import signals
from django.core.exceptions import ObjectDoesNotExist


class DiscussionManager(models.Manager):
    """ Manager for the :class:`Message` model. """

    def send_message(self, sender, to_user_list, subject, body):
        """
        Send a message from a user, to a user.

        :param sender:
            The :class:`User` which sends the message.

        :param to_user_list:
            A list which elements are :class:`User` to whom the message is for.

        :param message:
            String containing the message.
        :return:
            A Discussion :class:`Discussion`

        """
        discussion = self.model(sender=sender, subject=subject)
        discussion.save()

        discussion.save_recipients(list(to_user_list) + [sender])

        discussion.add_message(body)

        return discussion

    def get_conversation_between(self, from_user, to_user):
        """ Returns a conversation between two users """
        messages = self.filter(
            Q(sender=from_user, recipients=to_user, sender_deleted_at__isnull=True)
            | Q(
                sender=to_user,
                recipients=from_user,
                messagerecipient__deleted_at__isnull=True,
            )
        )
        return messages


class RecipientManager(models.Manager):
    """ Manager for the :class:`MessageRecipient` model. """

    def contribute_to_class(self, cls, name):
        signals.post_save.connect(self.handle_post_save, sender=cls)
        signals.post_delete.connect(self.handle_post_delete, sender=cls)
        return super(RecipientManager, self).contribute_to_class(cls, name)

    def handle_post_save(self, instance, **kwargs):
        if kwargs.get("created", False):
            discussion = instance.discussion
            discussion.update_counters()

    def handle_post_delete(self, instance, **kwargs):
        try:
            instance.discussion.update_counters()
        except ObjectDoesNotExist:
            pass

    def count_unread_messages_for(self, user):
        """
        Returns the amount of unread messages for this user

        :param user:
            A Django :class:`User`

        :return:
            An integer with the amount of unread messages.

        """
        unread_total = self.filter(user=user, status=self.model.STATUS.unread).count()

        return unread_total

    def count_unread_messages_between(self, to_user, from_user):
        """
        Returns the amount of unread messages between two users

        :param to_user:
            A Django :class:`User` for who the messages are for.

        :param from_user:
            A Django :class:`User` from whom the messages originate from.

        :return:
            An integer with the amount of unread messages.

        """
        unread_total = self.filter(
            discussion__sender=from_user, user=to_user, status=self.model.STATUS.unread
        ).count()

        return unread_total


class MessageManager(models.Manager):
    def contribute_to_class(self, cls, name):
        signals.post_save.connect(self.post_save, sender=cls)
        signals.post_delete.connect(self.post_delete, sender=cls)
        return super(MessageManager, self).contribute_to_class(cls, name)

    def post_save(self, instance, **kwargs):
        if kwargs.get("created", False):
            discussion = instance.discussion
            discussion.latest_message = instance
            discussion.update_counters()

    def post_delete(self, instance, **kwargs):
        try:
            instance.discussion.update_counters()
        except ObjectDoesNotExist:
            pass



from __future__ import unicode_literals

from django.test import TestCase

from ..forms import ComposeForm, ReplyForm, FolderForm
from ..models import Discussion
from ..compat import User


class ComposeFormTests(TestCase):
    """ Test the compose form. """

    fixtures = ["users.json"]

    def test_invalid_data(self):
        """
        Test the save method of :class:`ComposeForm`

        We don't need to make the ``to`` field sweat because we have done that
        in the ``fields`` test.

        """
        invalid_data_dicts = [
            # No body
            {
                "data": {"to": "john", "body": "", "subject": ""},
                "error": ("body", ["This field is required."]),
            }
        ]

        for invalid_dict in invalid_data_dicts:
            form = ComposeForm(data=invalid_dict["data"])
            assert form.is_valid() is False
            self.assertEqual(
                form.errors[invalid_dict["error"][0]], invalid_dict["error"][1]
            )

    def test_save_discussion(self):
        """ Test valid data """
        valid_data = {"to": "thoas,ampelmann", "body": "Body", "subject": "subject"}

        form = ComposeForm(data=valid_data)

        assert form.is_valid() is not False

        # Save the form.
        sender = User.objects.get(username="oleiade")
        discussion = form.save(sender)

        # Check if the values are set correctly
        self.assertEqual(discussion.subject, valid_data["subject"])

        self.assertEqual(discussion.messages.count(), 1)

        self.assertEqual(discussion.sender, sender)

        assert discussion.created_at is not None

        message = discussion.messages.all()[0]

        self.assertEqual(message.body, valid_data["body"])

        # Check recipients
        self.assertEqual(discussion.recipients.all()[0].username, "ampelmann")
        self.assertEqual(discussion.recipients.all()[1].username, "thoas")


class ReplyFormTests(TestCase):
    fixtures = ["users.json", "messages.json"]

    def test_save_message(self):
        valid_data = {"body": "Body"}

        discussion = Discussion.objects.get(pk=1)

        form = ReplyForm(discussion=discussion, data=valid_data)

        assert form.is_valid() is True

        sender = User.objects.get(username="oleiade")

        message = form.save(sender)

        self.assertEqual(message.body, valid_data["body"])

        self.assertEqual(discussion.messages.count(), 2)

        self.assertEqual(message.sender, sender)


class FolderFormTests(TestCase):
    fixtures = ["users.json"]

    def test_save_folder(self):
        valid_data = {"name": "My folder"}

        form = FolderForm(data=valid_data)

        assert form.is_valid() is True

        sender = User.objects.get(username="oleiade")

        folder = form.save(sender)

        self.assertEqual(folder.name, valid_data["name"])



import os

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "test_db",
        "USER": "",
        "PASSWORD": "",
        "HOST": "",
        "PORT": "",
    }
}

SITE_ID = 1
DEBUG = True

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "discussions",
]

ROOT_URLCONF = "discussions.urls"

TEMPLATE_DIRS = (os.path.join(os.path.dirname(__file__), "templates"),)

SECRET_KEY = "blabla"

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]
